// Copyright 2010 Jussi Lepisto

#include "Matrix2f.h"

#include <Eigen/Array>
#include <Eigen/Geometry>
#include <Eigen/LU>

#include "util/macros.h"
#include "vector/vector2f.h"

// Structures
PyGetSetDef Matrix2f_getset[] = {
	{"inverse",		(getter)Matrix2f_get_inverse,	NULL, "", NULL},
	{"transpose",	(getter)Matrix2f_get_transpose,
		(setter)Matrix2f_set_transpose, "", NULL},

	{NULL}
};

PyMethodDef Matrix2f_methods[] = {
	{"zero",		(PyCFunction)Matrix2f_zero,		METH_CLASS | METH_NOARGS,
	 ""},
	{"ones",		(PyCFunction)Matrix2f_ones,		METH_CLASS | METH_NOARGS,
	 ""},
	{"identity",	(PyCFunction)Matrix2f_identity,	METH_CLASS | METH_NOARGS,
	 ""},
	{"constant",	(PyCFunction)Matrix2f_constant,	METH_CLASS | METH_VARARGS,
	 ""},
	{"random",		(PyCFunction)Matrix2f_random,	METH_CLASS | METH_NOARGS,
	 ""},

	{"set",				(PyCFunction)Matrix2f_set,			METH_VARARGS,
	 ""},
	{"set_zero",		(PyCFunction)Matrix2f_set_zero,		METH_NOARGS,
	 ""},
	{"set_ones",		(PyCFunction)Matrix2f_set_ones,		METH_NOARGS,
	 ""},
	{"set_identity",	(PyCFunction)Matrix2f_set_identity,	METH_NOARGS,
	 ""},
	{"set_constant",	(PyCFunction)Matrix2f_set_constant,	METH_VARARGS,
	 ""},
	{"set_random",		(PyCFunction)Matrix2f_set_random,	METH_NOARGS,
	 ""},

	{"get_col",			(PyCFunction)Matrix2f_get_col,		METH_VARARGS,
	 ""},
	{"set_col",			(PyCFunction)Matrix2f_set_col,		METH_VARARGS,
	 ""},
	{"get_row",			(PyCFunction)Matrix2f_get_row,		METH_VARARGS,
	 ""},
	{"set_row",			(PyCFunction)Matrix2f_set_row,		METH_VARARGS,
	 ""},

	{"invert",			(PyCFunction)Matrix2f_invert,		METH_NOARGS,
	 ""},

	{NULL}
};

PyNumberMethods Matrix2f_numbermethods = {
	Matrix2f_add,				// nb_add
	Matrix2f_subtract,			// nb_subtract
	Matrix2f_multiply,			// nb_multiply
	Matrix2f_divide,			// nb_divide
	0,							// nb_remainder
	0,							// nb_divmod
	0,							// nb_power
	Matrix2f_negative,			// nb_negative
	0,							// nb_positive
	0,							// nb_absolute
	0,							// nb_nonzero
	0,							// nb_invert
	0,							// nb_lshift
	0,							// nb_rshift
	0,							// nb_and
	0,							// nb_xor
	0,							// nb_or
	0,							// nb_coerce
	0,							// nb_int
	0,							// nb_long
	0,							// nb_float
	0,							// nb_oct
	0,							// nb_hex
	Matrix2f_inplace_add,		// nb_inplace_add
	Matrix2f_inplace_subtract,	// nb_inplace_subtract
	Matrix2f_inplace_multiply,	// nb_inplace_multiply
	Matrix2f_inplace_divide,	// nb_inplace_divide
};

PyMappingMethods Matrix2f_mappingmethods = {
	(lenfunc)Matrix2f_length,				// mp_length
	(binaryfunc)Matrix2f_subscript,			// mp_subscript
	(objobjargproc)Matrix2f_ass_subscript,	// mp_ass_subscript
};

PyTypeObject Matrix2fType = {
	PyObject_HEAD_INIT(NULL)

	0,								// ob_size
	"pyeigen.Matrix2f",				// tp_name
	sizeof(Matrix2f),				// tp_basicsize
	0,								// tp_itemsize
	0,								// tp_dealloc
	0,								// tp_print
	0,								// tp_getattr
	0,								// tp_setattr
	0,								// tp_compare
	(reprfunc)Matrix2f_repr,		// tp_repr
	&Matrix2f_numbermethods,		// tp_as_number
	0,								// tp_as_sequence
	&Matrix2f_mappingmethods,		// tp_as_mapping
	0,								// tp_hash
	0,								// tp_call
	0,								// tp_str
	0,								// tp_getattro
	0,								// tp_setattro
	0,								// tp_as_buffer
	Py_TPFLAGS_DEFAULT |			// tp_flags
	Py_TPFLAGS_BASETYPE |
	Py_TPFLAGS_CHECKTYPES,
	0,								// tp_doc
	0,								// tp_traverse
	0,								// tp_clear
	0,								// tp_richcompare
	0,								// tp_weaklistoffset
	0,								// tp_iter
	0,								// tp_iternext
	Matrix2f_methods,				// tp_methods
	0,								// tp_members
	Matrix2f_getset,				// tp_getset
	0,								// tp_base
	0,								// tp_dict
	0,								// tp_descr_get
	0,								// tp_descr_set
	0,								// tp_dictoffset
	(initproc)Matrix2f_init,		// tp_init
	0,								// tp_alloc
	Matrix2f_new,					// tp_new
	0,								// tp_free
	0,								// tp_is_gc
	0,								// tp_bases
	0,								// tp_mro
	0,								// tp_cache
	0,								// tp_subclasses
	0,								// tp_weaklist
	0,								// tp_del
	0,								// tp_version_tag
};

// Helpers
int Matrix2f_Check(PyObject* p)
{
	return p->ob_type == &Matrix2fType;
}

int Matrix2f_ParseKey(PyObject* key, Py_ssize_t& key1, Py_ssize_t& key2)
{
	if(!PyArg_ParseTuple(key, "nn", &key1, &key2))
		return 0;

	if(key1 < 0)
		key1 += 2;
	if(key2 < 0)
		key2 += 2;

	if(key1 >= 2 || key2 >= 2)
	{
		PyErr_SetString(PyExc_IndexError, "index out of range");
		return 0;
	}

	return 1;
}

// Construction
PyObject* Matrix2f_new(PyTypeObject* type, PyObject* args, PyObject* kwds)
{
	return type->tp_alloc(type, 0);
}

int Matrix2f_init(Matrix2f* self, PyObject* args, PyObject* kwds)
{
	if(!Matrix2f_set(self, args))
		return -1;

	return 0;
}

// Properties
PyObject* Matrix2f_get_inverse(Matrix2f* self, void* closure)
{
	Matrix2f* result = PyObject_New(Matrix2f, &Matrix2fType);
	if(result != NULL)
		self->matrix.computeInverse(&result->matrix);

	return (PyObject*)result;
}

PyObject* Matrix2f_get_transpose(Matrix2f* self, void* closure)
{
	Matrix2f* result = PyObject_New(Matrix2f, &Matrix2fType);
	if(result != NULL)
		result->matrix = self->matrix.transpose();

	return (PyObject*)result;
}

int Matrix2f_set_transpose(Matrix2f* self, PyObject* value, void* closure)
{
	if(!Matrix2f_Check(value))
	{
		PyErr_SetString(PyExc_TypeError, "transpose must be Matrix2f");
		return -1;
	}

	Matrix2f* m = (Matrix2f*)value;
	self->matrix.transpose() = m->matrix;
	return 0;
}

// Methods
PyObject* Matrix2f_zero(PyTypeObject* cls, PyObject* noargs)
{
	Matrix2f* result = PyObject_New(Matrix2f, cls);
	if(result != NULL)
		result->matrix.setZero();

	return (PyObject*)result;
}

PyObject* Matrix2f_ones(PyTypeObject* cls, PyObject* noargs)
{
	Matrix2f* result = PyObject_New(Matrix2f, cls);
	if(result != NULL)
		result->matrix.setOnes();

	return (PyObject*)result;
}

PyObject* Matrix2f_identity(PyTypeObject* cls, PyObject* noargs)
{
	Matrix2f* result = PyObject_New(Matrix2f, cls);
	if(result != NULL)
		result->matrix.setIdentity();

	return (PyObject*)result;
}

PyObject* Matrix2f_constant(PyTypeObject* cls, PyObject* args)
{
	float constant = 0.0f;
	if(!PyArg_ParseTuple(args, "f", &constant))
		return NULL;

	Matrix2f* result = PyObject_New(Matrix2f, cls);
	if(result != NULL)
		result->matrix.setConstant(constant);

	return (PyObject*)result;
}

PyObject* Matrix2f_random(PyTypeObject* cls, PyObject* noargs)
{
	Matrix2f* result = PyObject_New(Matrix2f, cls);
	if(result != NULL)
		result->matrix.setRandom();

	return (PyObject*)result;
}

PyObject* Matrix2f_set(Matrix2f* self, PyObject* args)
{
	float m11 = 0.0f, m12 = 0.0f;
	float m21 = 0.0f, m22 = 0.0f;
	if(!PyArg_ParseTuple(args, "|ffff", &m11, &m12, &m21, &m22))
		return NULL;

	self->matrix << m11, m12, m21, m22;
	Py_RETURN_NONE;
}

PyObject* Matrix2f_set_zero(Matrix2f* self, PyObject* noargs)
{
	self->matrix.setZero();
	Py_RETURN_NONE;
}

PyObject* Matrix2f_set_ones(Matrix2f* self, PyObject* noargs)
{
	self->matrix.setOnes();
	Py_RETURN_NONE;
}

PyObject* Matrix2f_set_identity(Matrix2f* self, PyObject* noargs)
{
	self->matrix.setIdentity();
	Py_RETURN_NONE;
}

PyObject* Matrix2f_set_constant(Matrix2f* self, PyObject* args)
{
	float constant = 0.0f;
	if(!PyArg_ParseTuple(args, "f", &constant))
		return NULL;

	self->matrix.setConstant(constant);
	Py_RETURN_NONE;
}

PyObject* Matrix2f_set_random(Matrix2f* self, PyObject* noargs)
{
	self->matrix.setRandom();
	Py_RETURN_NONE;
}

PyObject* Matrix2f_get_col(Matrix2f* self, PyObject* args)
{
	Py_ssize_t index;
	if(!PyArg_ParseTuple(args, "n", &index))
		return NULL;

	if(index < 0 || index >= 2)
	{
		PyErr_SetString(PyExc_IndexError, "index out of range");
		return NULL;
	}

	Vector2f* result = PyObject_New(Vector2f, &Vector2fType);
	if(result != NULL)
		result->vector = self->matrix.col(index);

	return (PyObject*)result;
}

PyObject* Matrix2f_set_col(Matrix2f* self, PyObject* args)
{
	Py_ssize_t index;
	Vector2f* value = NULL;
	if(!PyArg_ParseTuple(args, "nO!", &index, &Vector2fType, &value))
		return NULL;

	if(index < 0 || index >= 2)
	{
		PyErr_SetString(PyExc_IndexError, "index out of range");
		return NULL;
	}

	self->matrix.col(index) = value->vector;
	Py_RETURN_NONE;
}

PyObject* Matrix2f_get_row(Matrix2f* self, PyObject* args)
{
	Py_ssize_t index;
	if(!PyArg_ParseTuple(args, "n", &index))
		return NULL;

	if(index < 0 || index >= 2)
	{
		PyErr_SetString(PyExc_IndexError, "index out of range");
		return NULL;
	}

	Vector2f* result = PyObject_New(Vector2f, &Vector2fType);
	if(result != NULL)
		result->vector = self->matrix.row(index);

	return (PyObject*)result;
}

PyObject* Matrix2f_set_row(Matrix2f* self, PyObject* args)
{
	Py_ssize_t index;
	Vector2f* value = NULL;
	if(!PyArg_ParseTuple(args, "nO!", &index, &Vector2fType, &value))
		return NULL;

	if(index < 0 || index >= 2)
	{
		PyErr_SetString(PyExc_IndexError, "index out of range");
		return NULL;
	}

	self->matrix.row(index) = value->vector;
	Py_RETURN_NONE;
}

PyObject* Matrix2f_invert(Matrix2f* self, PyObject* noargs)
{
	Eigen::Matrix2f temp;
	self->matrix.computeInverse(&temp);
	self->matrix = temp;
	Py_RETURN_NONE;
}

// Number methods
PyObject* Matrix2f_add(PyObject* o1, PyObject* o2)
{
	if(!Matrix2f_Check(o1) || !Matrix2f_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Matrix2f* m1 = (Matrix2f*)o1;
	Matrix2f* m2 = (Matrix2f*)o2;
	Matrix2f* result = PyObject_New(Matrix2f, &Matrix2fType);
	if(result != NULL)
		result->matrix = m1->matrix + m2->matrix;

	return (PyObject*)result;
}

PyObject* Matrix2f_subtract(PyObject* o1, PyObject* o2)
{
	if(!Matrix2f_Check(o1) || !Matrix2f_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Matrix2f* m1 = (Matrix2f*)o1;
	Matrix2f* m2 = (Matrix2f*)o2;
	Matrix2f* result = PyObject_New(Matrix2f, &Matrix2fType);
	if(result != NULL)
		result->matrix = m1->matrix - m2->matrix;

	return (PyObject*)result;
}

PyObject* Matrix2f_multiply(PyObject* o1, PyObject* o2)
{
	// Matrix multiply
	if(Matrix2f_Check(o1) && Matrix2f_Check(o2))
	{
		Matrix2f* m1 = (Matrix2f*)o1;
		Matrix2f* m2 = (Matrix2f*)o2;
		Matrix2f* result = PyObject_New(Matrix2f, &Matrix2fType);
		if(result != NULL)
			result->matrix = m1->matrix * m2->matrix;

		return (PyObject*)result;
	}
	// Vector multiply
	else if(Matrix2f_Check(o1) && Vector2f_Check(o2))
	{
		Matrix2f* m = (Matrix2f*)o1;
		Vector2f* v = (Vector2f*)o2;
		if(PyErr_Occurred() != NULL)
			return NULL;

		Vector2f* result = PyObject_New(Vector2f, &Vector2fType);
		if(result != NULL)
			result->vector = m->matrix * v->vector;

		return (PyObject*)result;
	}
	// Scalar multiply
	else if(Matrix2f_Check(o1) && PyNumber_Check(o2))
	{
		Matrix2f* m = (Matrix2f*)o1;
		float scalar = (float)PyFloat_AsDouble(o2);
		if(PyErr_Occurred() != NULL)
			return NULL;

		Matrix2f* result = PyObject_New(Matrix2f, &Matrix2fType);
		if(result != NULL)
			result->matrix = m->matrix * scalar;

		return (PyObject*)result;
	}
	else if(PyNumber_Check(o1) && Matrix2f_Check(o2))
	{
		float scalar = (float)PyFloat_AsDouble(o1);
		Matrix2f* m = (Matrix2f*)o2;
		if(PyErr_Occurred() != NULL)
			return NULL;

		Matrix2f* result = PyObject_New(Matrix2f, &Matrix2fType);
		if(result != NULL)
			result->matrix = m->matrix * scalar;

		return (PyObject*)result;
	}

	RETURN_NOTIMPLEMENTED;
}

PyObject* Matrix2f_divide(PyObject* o1, PyObject* o2)
{
	if(!Matrix2f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Matrix2f* m = (Matrix2f*)o1;
	float scalar = (float)PyFloat_AsDouble(o2);
	if(PyErr_Occurred() != NULL)
		return NULL;

	Matrix2f* result = PyObject_New(Matrix2f, &Matrix2fType);
	if(result != NULL)
		result->matrix = m->matrix / scalar;

	return (PyObject*)result;
}

PyObject* Matrix2f_negative(PyObject* o)
{
	if(!Matrix2f_Check(o))
		RETURN_NOTIMPLEMENTED;

	Matrix2f* m = (Matrix2f*)o;
	Matrix2f* result = PyObject_New(Matrix2f, &Matrix2fType);
	if(result != NULL)
		result->matrix = -m->matrix;

	return (PyObject*)result;
}

PyObject* Matrix2f_inplace_add(PyObject* o1, PyObject* o2)
{
	if(!Matrix2f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Matrix2f* m1 = (Matrix2f*)o1;
	Matrix2f* m2 = (Matrix2f*)o2;
	m1->matrix += m2->matrix;

	return o1;
}

PyObject* Matrix2f_inplace_subtract(PyObject* o1, PyObject* o2)
{
	if(!Matrix2f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Matrix2f* m1 = (Matrix2f*)o1;
	Matrix2f* m2 = (Matrix2f*)o2;
	m1->matrix -= m2->matrix;

	return o1;
}

PyObject* Matrix2f_inplace_multiply(PyObject* o1, PyObject* o2)
{
	Matrix2f* m1 = (Matrix2f*)o1;

	// Matrix multiply
	if(Matrix2f_Check(o2))
	{
		Matrix2f* m2 = (Matrix2f*)o2;
		m1->matrix *= m2->matrix;
	}
	// Scalar multiply
	else if(PyNumber_Check(o2))
	{
		float scalar = (float)PyFloat_AsDouble(o2);
		if(PyErr_Occurred() != NULL)
			return NULL;

		m1->matrix *= scalar;
	}
	else
		RETURN_NOTIMPLEMENTED;

	return (PyObject*)m1;
}

PyObject* Matrix2f_inplace_divide(PyObject* o1, PyObject* o2)
{
	if(!Matrix2f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Matrix2f* m = (Matrix2f*)o1;
	float scalar = (float)PyFloat_AsDouble(o2);
	if(PyErr_Occurred() != NULL)
		return NULL;

	m->matrix /= scalar;

	return (PyObject*)m;
}

// Mapping methods
Py_ssize_t Matrix2f_length(Matrix2f* self)
{
	return 4;
}

PyObject* Matrix2f_subscript(Matrix2f* self, PyObject* key)
{
	Py_ssize_t key1, key2;
	if(!Matrix2f_ParseKey(key, key1, key2))
		return NULL;

	return PyFloat_FromDouble(self->matrix(key1, key2));
}

int Matrix2f_ass_subscript(Matrix2f* self, PyObject* key, PyObject* value)
{
	Py_ssize_t key1, key2;
	if(!Matrix2f_ParseKey(key, key1, key2))
		return -1;

	self->matrix(key1, key2) = (float)PyFloat_AsDouble(value);
	return 0;
}

// Special methods
PyObject* Matrix2f_repr(Matrix2f* self)
{
	char buffer[1024];
	_snprintf(buffer, 1024, "Matrix2f(%f, %f,\n         %f, %f)",
		self->matrix[0], self->matrix[1], self->matrix[2], self->matrix[3]);
	return PyString_FromString(buffer);
}
