// Copyright 2010 Jussi Lepisto

#include "vector3f.h"

#include <Eigen/Array>
#include <Eigen/Geometry>

#include "util/macros.h"


// Structures
PyGetSetDef Vector3f_getset[] = {
	{"x", (getter)Vector3f_get_x, (setter)Vector3f_set_x, "", NULL},
	{"y", (getter)Vector3f_get_y, (setter)Vector3f_set_y, "", NULL},
	{"z", (getter)Vector3f_get_z, (setter)Vector3f_set_z, "", NULL},

	{"norm",		(getter)Vector3f_get_norm,			NULL, "", NULL},
	{"norm_sq",		(getter)Vector3f_get_norm_sq,		NULL, "", NULL},
	{"normalized",	(getter)Vector3f_get_normalized,	NULL, "", NULL},

	{NULL}
};

PyMethodDef Vector3f_methods[] = {
	{"zero",		(PyCFunction)Vector3f_zero,		METH_CLASS | METH_NOARGS,
	 ""},
	{"ones",		(PyCFunction)Vector3f_ones,		METH_CLASS | METH_NOARGS,
	 ""},
	{"identity",	(PyCFunction)Vector3f_identity,	METH_CLASS | METH_NOARGS,
	 ""},
	{"constant",	(PyCFunction)Vector3f_constant,	METH_CLASS | METH_VARARGS,
	 ""},
	{"random",		(PyCFunction)Vector3f_random,	METH_CLASS | METH_NOARGS,
	 ""},
	{"unit_x",		(PyCFunction)Vector3f_unit_x,	METH_CLASS | METH_NOARGS,
	 ""},
	{"unit_y",		(PyCFunction)Vector3f_unit_y,	METH_CLASS | METH_NOARGS,
	 ""},
	{"unit_z",		(PyCFunction)Vector3f_unit_z,	METH_CLASS | METH_NOARGS,
	 ""},

	{"set",				(PyCFunction)Vector3f_set,			METH_VARARGS,
	 ""},
	{"set_zero",		(PyCFunction)Vector3f_set_zero,		METH_NOARGS,
	 ""},
	{"set_ones",		(PyCFunction)Vector3f_set_ones,		METH_NOARGS,
	 ""},
	{"set_constant",	(PyCFunction)Vector3f_set_constant,	METH_VARARGS,
	 ""},
	{"set_random",		(PyCFunction)Vector3f_set_random,	METH_NOARGS,
	 ""},

	{"dot",				(PyCFunction)Vector3f_dot,			METH_O,
	 ""},
	{"cross",			(PyCFunction)Vector3f_cross,		METH_O,
	 ""},

	{"normalize",		(PyCFunction)Vector3f_normalize,	METH_NOARGS,
	 ""},

	{NULL}
};

PyNumberMethods Vector3f_numbermethods = {
	Vector3f_add,				// nb_add
	Vector3f_subtract,			// nb_subtract
	Vector3f_multiply,			// nb_multiply
	Vector3f_divide,			// nb_divide
	0,							// nb_remainder
	0,							// nb_divmod
	0,							// nb_power
	Vector3f_negative,			// nb_negative
	0,							// nb_positive
	0,							// nb_absolute
	0,							// nb_nonzero
	0,							// nb_invert
	0,							// nb_lshift
	0,							// nb_rshift
	0,							// nb_and
	0,							// nb_xor
	0,							// nb_or
	0,							// nb_coerce
	0,							// nb_int
	0,							// nb_long
	0,							// nb_float
	0,							// nb_oct
	0,							// nb_hex
	Vector3f_inplace_add,		// nb_inplace_add
	Vector3f_inplace_subtract,	// nb_inplace_subtract
	Vector3f_inplace_multiply,	// nb_inplace_multiply
	Vector3f_inplace_divide,	// nb_inplace_divide
};

PySequenceMethods Vector3f_sequencemethods = {
	(lenfunc)Vector3f_length,					// sq_length
	0,											// sq_concat
	0,											// sq_repeat
	(ssizeargfunc)Vector3f_item,				// sq_item
	(ssizessizeargfunc)Vector3f_slice,			// sq_slice
	(ssizeobjargproc)Vector3f_ass_item,			// sq_ass_item
	(ssizessizeobjargproc)Vector3f_ass_slice,	// sq_ass_slice
};

PyTypeObject Vector3fType = {
	PyObject_HEAD_INIT(NULL)

	0,								// ob_size
	"pyeigen.Vector3f",				// tp_name
	sizeof(Vector3f),				// tp_basicsize
	0,								// tp_itemsize
	0,								// tp_dealloc
	0,								// tp_print
	0,								// tp_getattr
	0,								// tp_setattr
	0,								// tp_compare
	(reprfunc)Vector3f_repr,		// tp_repr
	&Vector3f_numbermethods,		// tp_as_number
	&Vector3f_sequencemethods,		// tp_as_sequence
	0,								// tp_as_mapping
	0,								// tp_hash
	0,								// tp_call
	0,								// tp_str
	0,								// tp_getattro
	0,								// tp_setattro
	0,								// tp_as_buffer
	Py_TPFLAGS_DEFAULT |			// tp_flags
	Py_TPFLAGS_BASETYPE |
	Py_TPFLAGS_CHECKTYPES,
	0,								// tp_doc
	0,								// tp_traverse
	0,								// tp_clear
	0,								// tp_richcompare
	0,								// tp_weaklistoffset
	0,								// tp_iter
	0,								// tp_iternext
	Vector3f_methods,				// tp_methods
	0,								// tp_members
	Vector3f_getset,				// tp_getset
	0,								// tp_base
	0,								// tp_dict
	0,								// tp_descr_get
	0,								// tp_descr_set
	0,								// tp_dictoffset
	(initproc)Vector3f_init,		// tp_init
	0,								// tp_alloc
	Vector3f_new,					// tp_new
	0,								// tp_free
	0,								// tp_is_gc
	0,								// tp_bases
	0,								// tp_mro
	0,								// tp_cache
	0,								// tp_subclasses
	0,								// tp_weaklist
	0,								// tp_del
	0,								// tp_version_tag
};

// Helpers
int Vector3f_Check(PyObject* p)
{
	return p->ob_type == &Vector3fType;
}

// Construction
PyObject* Vector3f_new(PyTypeObject* type, PyObject* args, PyObject* kwds)
{
	return type->tp_alloc(type, 0);
}

int Vector3f_init(Vector3f* self, PyObject* args, PyObject* kwds)
{
	if(!Vector3f_set(self, args))
		return -1;

	return 0;
}

// Properties
PyObject* Vector3f_get_x(Vector3f* self, void* closure)
{
	return PyFloat_FromDouble(self->vector.x());
}

int Vector3f_set_x(Vector3f* self, PyObject* value, void* closure)
{
	self->vector.x() = (float)PyFloat_AsDouble(value);
	if(PyErr_Occurred() != NULL)
		return -1;

	return 0;
}

PyObject* Vector3f_get_y(Vector3f* self, void* closure)
{
	return PyFloat_FromDouble(self->vector.y());
}

int Vector3f_set_y(Vector3f* self, PyObject* value, void* closure)
{
	self->vector.y() = (float)PyFloat_AsDouble(value);
	if(PyErr_Occurred() != NULL)
		return -1;

	return 0;
}

PyObject* Vector3f_get_z(Vector3f* self, void* closure)
{
	return PyFloat_FromDouble(self->vector.z());
}

int Vector3f_set_z(Vector3f* self, PyObject* value, void* closure)
{
	self->vector.z() = (float)PyFloat_AsDouble(value);
	if(PyErr_Occurred() != NULL)
		return -1;

	return 0;
}

PyObject* Vector3f_get_norm(Vector3f* self, void* closure)
{
	return PyFloat_FromDouble(self->vector.norm());
}

PyObject* Vector3f_get_norm_sq(Vector3f* self, void* closure)
{
	return PyFloat_FromDouble(self->vector.squaredNorm());
}

PyObject* Vector3f_get_normalized(Vector3f* self, void* closure)
{
	Vector3f* result = PyObject_New(Vector3f, &Vector3fType);
	if(result != NULL)
		result->vector = self->vector.normalized();

	return (PyObject*)result;
}

// Methods
PyObject* Vector3f_zero(PyTypeObject* cls, PyObject* noargs)
{
	Vector3f* result = PyObject_New(Vector3f, cls);
	if(result != NULL)
		result->vector.setZero();

	return (PyObject*)result;
}

PyObject* Vector3f_ones(PyTypeObject* cls, PyObject* noargs)
{
	Vector3f* result = PyObject_New(Vector3f, cls);
	if(result != NULL)
		result->vector.setOnes();

	return (PyObject*)result;
}

PyObject* Vector3f_identity(PyTypeObject* cls, PyObject* noargs)
{
	Vector3f* result = PyObject_New(Vector3f, cls);
	if(result != NULL)
		result->vector.setIdentity();

	return (PyObject*)result;
}

PyObject* Vector3f_constant(PyTypeObject* cls, PyObject* args)
{
	float constant = 0.0f;
	if(!PyArg_ParseTuple(args, "f", &constant))
		return NULL;

	Vector3f* result = PyObject_New(Vector3f, cls);
	if(result != NULL)
		result->vector.setConstant(constant);

	return (PyObject*)result;
}

PyObject* Vector3f_random(PyTypeObject* cls, PyObject* noargs)
{
	Vector3f* result = PyObject_New(Vector3f, cls);
	if(result != NULL)
		result->vector.setRandom();

	return (PyObject*)result;
}

PyObject* Vector3f_unit_x(PyTypeObject* cls, PyObject* noargs)
{
	Vector3f* result = PyObject_New(Vector3f, cls);
	if(result != NULL)
		result->vector << 1.f, 0.f, 0.f;

	return (PyObject*)result;
}

PyObject* Vector3f_unit_y(PyTypeObject* cls, PyObject* noargs)
{
	Vector3f* result = PyObject_New(Vector3f, cls);
	if(result != NULL)
		result->vector << 0.f, 1.f, 0.f;

	return (PyObject*)result;
}

PyObject* Vector3f_unit_z(PyTypeObject* cls, PyObject* noargs)
{
	Vector3f* result = PyObject_New(Vector3f, cls);
	if(result != NULL)
		result->vector << 0.f, 0.f, 1.f;

	return (PyObject*)result;
}

PyObject* Vector3f_set(Vector3f* self, PyObject* args)
{
	float x = 0.0f;
	float y = 0.0f;
	float z = 0.0f;
	if(!PyArg_ParseTuple(args, "|fff", &x, &y, &z))
		return NULL;

	self->vector << x, y, z;
	Py_RETURN_NONE;
}

PyObject* Vector3f_set_zero(Vector3f* self, PyObject* noargs)
{
	self->vector.setZero();
	Py_RETURN_NONE;
}

PyObject* Vector3f_set_ones(Vector3f* self, PyObject* noargs)
{
	self->vector.setOnes();
	Py_RETURN_NONE;
}

PyObject* Vector3f_set_constant(Vector3f* self, PyObject* args)
{
	float constant = 0.0f;
	if(!PyArg_ParseTuple(args, "f", &constant))
		return NULL;

	self->vector.setConstant(constant);
	Py_RETURN_NONE;
}

PyObject* Vector3f_set_random(Vector3f* self, PyObject* noargs)
{
	self->vector.setRandom();
	Py_RETURN_NONE;
}

PyObject* Vector3f_dot(Vector3f* self, PyObject* other)
{
	if(!Vector3f_Check(other))
	{
		PyErr_SetString(PyExc_ValueError, "Need Vector3f for dot product");
		return NULL;
	}

	Vector3f* otherv = (Vector3f*)other;
	return PyFloat_FromDouble(self->vector.dot(otherv->vector));
}

PyObject* Vector3f_cross(Vector3f* self, PyObject* other)
{
	if(!Vector3f_Check(other))
	{
		PyErr_SetString(PyExc_ValueError, "Need Vector3f for cross product");
		return NULL;
	}

	Vector3f* otherv = (Vector3f*)other;
	Vector3f* result = PyObject_New(Vector3f, &Vector3fType);
	if(result != NULL)
		result->vector = self->vector.cross(otherv->vector);

	return (PyObject*)result;
}

PyObject* Vector3f_normalize(Vector3f* self, PyObject* noargs)
{
	self->vector.normalize();
	Py_RETURN_NONE;
}

// Number methods
PyObject* Vector3f_add(PyObject* o1, PyObject* o2)
{
	if(!Vector3f_Check(o1) || !Vector3f_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector3f* v1 = (Vector3f*)o1;
	Vector3f* v2 = (Vector3f*)o2;
	Vector3f* result = PyObject_New(Vector3f, &Vector3fType);
	if(result != NULL)
		result->vector = v1->vector + v2->vector;

	return (PyObject*)result;
}

PyObject* Vector3f_subtract(PyObject* o1, PyObject* o2)
{
	if(!Vector3f_Check(o1) || !Vector3f_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector3f* v1 = (Vector3f*)o1;
	Vector3f* v2 = (Vector3f*)o2;
	Vector3f* result = PyObject_New(Vector3f, &Vector3fType);
	if(result != NULL)
		result->vector = v1->vector - v2->vector;

	return (PyObject*)result;
}

PyObject* Vector3f_multiply(PyObject* o1, PyObject* o2)
{
	Vector3f* v;
	float scalar;
	if(Vector3f_Check(o1) && PyNumber_Check(o2))
	{
		v = (Vector3f*)o1;
		scalar = (float)PyFloat_AsDouble(o2);
		if(PyErr_Occurred() != NULL)
			return NULL;

	}
	else if(PyNumber_Check(o1) && Vector3f_Check(o2))
	{
		scalar = (float)PyFloat_AsDouble(o1);
		v = (Vector3f*)o2;
		if(PyErr_Occurred() != NULL)
			return NULL;
	}
	else
		RETURN_NOTIMPLEMENTED;

	Vector3f* result = PyObject_New(Vector3f, &Vector3fType);
	if(result != NULL)
		result->vector = v->vector * scalar;

	return (PyObject*)result;
}

PyObject* Vector3f_divide(PyObject* o1, PyObject* o2)
{
	if(!Vector3f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector3f* v = (Vector3f*)o1;
	float scalar = (float)PyFloat_AsDouble(o2);
	if(PyErr_Occurred() != NULL)
		return NULL;

	Vector3f* result = PyObject_New(Vector3f, &Vector3fType);
	if(result != NULL)
		result->vector = v->vector / scalar;

	return (PyObject*)result;
}

PyObject* Vector3f_negative(PyObject* o)
{
	if(!Vector3f_Check(o))
		RETURN_NOTIMPLEMENTED;

	Vector3f* v = (Vector3f*)o;
	Vector3f* result = PyObject_New(Vector3f, &Vector3fType);
	if(result != NULL)
		result->vector = -v->vector;

	return (PyObject*)result;
}

PyObject* Vector3f_inplace_add(PyObject* o1, PyObject* o2)
{
	if(!Vector3f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector3f* v1 = (Vector3f*)o1;
	Vector3f* v2 = (Vector3f*)o2;
	v1->vector += v2->vector;

	return o1;
}

PyObject* Vector3f_inplace_subtract(PyObject* o1, PyObject* o2)
{
	if(!Vector3f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector3f* v1 = (Vector3f*)o1;
	Vector3f* v2 = (Vector3f*)o2;
	v1->vector -= v2->vector;

	return o1;
}

PyObject* Vector3f_inplace_multiply(PyObject* o1, PyObject* o2)
{
	Vector3f* v;
	float scalar;
	if(Vector3f_Check(o1) && PyNumber_Check(o2))
	{
		v = (Vector3f*)o1;
		scalar = (float)PyFloat_AsDouble(o2);
		if(PyErr_Occurred() != NULL)
			return NULL;
	}
	else if(PyNumber_Check(o1) && Vector3f_Check(o2))
	{
		scalar = (float)PyFloat_AsDouble(o1);
		v = (Vector3f*)o2;
		if(PyErr_Occurred() != NULL)
			return NULL;
	}
	else
		RETURN_NOTIMPLEMENTED;

	v->vector *= scalar;

	return (PyObject*)v;
}

PyObject* Vector3f_inplace_divide(PyObject* o1, PyObject* o2)
{
	if(!Vector3f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector3f* v = (Vector3f*)o1;
	float scalar = (float)PyFloat_AsDouble(o2);
	if(PyErr_Occurred() != NULL)
		return NULL;

	v->vector /= scalar;

	return (PyObject*)v;
}
// Sequence methods
Py_ssize_t Vector3f_length(Vector3f* self)
{
	return 3;
}

PyObject* Vector3f_item(Vector3f* self, Py_ssize_t index)
{
	if(index < 0 || index >= 3)
	{
		PyErr_SetString(PyExc_IndexError, "index out of range");
		return NULL;
	}

	return PyFloat_FromDouble(self->vector[index]);
}

PyObject* Vector3f_slice(Vector3f* self, Py_ssize_t index1, Py_ssize_t index2)
{
	if(index1 > 3)
		index1 = 3;
	if(index2 > 3)
		index2 = 3;
	if(index1 > index2)
		index1 = index2;

	Py_ssize_t length = index2 - index1;
	PyObject* tuple = PyTuple_New(length);
	if(tuple != NULL)
		for(Py_ssize_t index = 0; index < length; ++index)
		{
			PyObject* item = PyFloat_FromDouble(self->vector[index1 + index]);
			if(item == NULL)
			{
				Py_DECREF(tuple);
				return NULL;
			}

			PyTuple_SetItem(tuple, index, item);
		}

	return tuple;
}

int Vector3f_ass_item(Vector3f* self, Py_ssize_t index, PyObject* value)
{
	if(index < 0 || index >= 3)
	{
		PyErr_SetString(PyExc_IndexError, "index out of range");
		return NULL;
	}

	self->vector[index] = (float)PyFloat_AsDouble(value);
	if(PyErr_Occurred() != NULL)
		return -1;

	return 0;
}

int Vector3f_ass_slice(Vector3f* self, Py_ssize_t index1, Py_ssize_t index2,
	PyObject* value)
{
	if(index1 > 3)
		index1 = 3;
	if(index2 > 3)
		index2 = 3;
	if(index1 > index2)
		index1 = index2;

	Py_ssize_t length = index2 - index1;

	if(value == NULL || PySequence_Length(value) != length)
	{
		PyErr_SetString(PyExc_ValueError, "Can't change size of Vector3f");
		return -1;
	}

	for(Py_ssize_t index = index1; index < index2; ++index)
	{
		self->vector[index] =
			(float)PyFloat_AsDouble(PySequence_GetItem(value, index));
		if(PyErr_Occurred() != NULL)
			return -1;
	}

	return 0;
}

// Special methods
PyObject* Vector3f_repr(Vector3f* self)
{
	char buffer[1024];
	_snprintf(buffer, 1024, "Vector3f(%f, %f, %f)", self->vector.x(),
		self->vector.y(), self->vector.z());
	return PyString_FromString(buffer);
}
