// Copyright 2010 Jussi Lepisto

#include "vector4f.h"

#include <Eigen/Array>
#include <Eigen/Geometry>

#include "util/macros.h"


// Structures
PyGetSetDef Vector4f_getset[] = {
	{"x", (getter)Vector4f_get_x, (setter)Vector4f_set_x, "", NULL},
	{"y", (getter)Vector4f_get_y, (setter)Vector4f_set_y, "", NULL},
	{"z", (getter)Vector4f_get_z, (setter)Vector4f_set_z, "", NULL},
	{"w", (getter)Vector4f_get_w, (setter)Vector4f_set_w, "", NULL},

	{"norm",		(getter)Vector4f_get_norm,			NULL, "", NULL},
	{"norm_sq",		(getter)Vector4f_get_norm_sq,		NULL, "", NULL},
	{"normalized",	(getter)Vector4f_get_normalized,	NULL, "", NULL},

	{NULL}
};

PyMethodDef Vector4f_methods[] = {
	{"zero",		(PyCFunction)Vector4f_zero,		METH_CLASS | METH_NOARGS,
	 ""},
	{"ones",		(PyCFunction)Vector4f_ones,		METH_CLASS | METH_NOARGS,
	 ""},
	{"identity",	(PyCFunction)Vector4f_identity,	METH_CLASS | METH_NOARGS,
	 ""},
	{"constant",	(PyCFunction)Vector4f_constant,	METH_CLASS | METH_VARARGS,
	 ""},
	{"random",		(PyCFunction)Vector4f_random,	METH_CLASS | METH_NOARGS,
	 ""},
	{"unit_x",		(PyCFunction)Vector4f_unit_x,	METH_CLASS | METH_NOARGS,
	 ""},
	{"unit_y",		(PyCFunction)Vector4f_unit_y,	METH_CLASS | METH_NOARGS,
	 ""},
	{"unit_z",		(PyCFunction)Vector4f_unit_z,	METH_CLASS | METH_NOARGS,
	 ""},
	{"unit_w",		(PyCFunction)Vector4f_unit_w,	METH_CLASS | METH_NOARGS,
	 ""},

	{"set",				(PyCFunction)Vector4f_set,			METH_VARARGS,
	 ""},
	{"set_zero",		(PyCFunction)Vector4f_set_zero,		METH_NOARGS,
	 ""},
	{"set_ones",		(PyCFunction)Vector4f_set_ones,		METH_NOARGS,
	 ""},
	{"set_constant",	(PyCFunction)Vector4f_set_constant,	METH_VARARGS,
	 ""},
	{"set_random",		(PyCFunction)Vector4f_set_random,	METH_NOARGS,
	 ""},

	{"dot",				(PyCFunction)Vector4f_dot,			METH_O,
	 ""},

	{"normalize",		(PyCFunction)Vector4f_normalize,	METH_NOARGS,
	 ""},

	{NULL}
};

PyNumberMethods Vector4f_numbermethods = {
	Vector4f_add,				// nb_add
	Vector4f_subtract,			// nb_subtract
	Vector4f_multiply,			// nb_multiply
	Vector4f_divide,			// nb_divide
	0,							// nb_remainder
	0,							// nb_divmod
	0,							// nb_power
	Vector4f_negative,			// nb_negative
	0,							// nb_positive
	0,							// nb_absolute
	0,							// nb_nonzero
	0,							// nb_invert
	0,							// nb_lshift
	0,							// nb_rshift
	0,							// nb_and
	0,							// nb_xor
	0,							// nb_or
	0,							// nb_coerce
	0,							// nb_int
	0,							// nb_long
	0,							// nb_float
	0,							// nb_oct
	0,							// nb_hex
	Vector4f_inplace_add,		// nb_inplace_add
	Vector4f_inplace_subtract,	// nb_inplace_subtract
	Vector4f_inplace_multiply,	// nb_inplace_multiply
	Vector4f_inplace_divide,	// nb_inplace_divide
};

PySequenceMethods Vector4f_sequencemethods = {
	(lenfunc)Vector4f_length,					// sq_length
	0,											// sq_concat
	0,											// sq_repeat
	(ssizeargfunc)Vector4f_item,				// sq_item
	(ssizessizeargfunc)Vector4f_slice,			// sq_slice
	(ssizeobjargproc)Vector4f_ass_item,			// sq_ass_item
	(ssizessizeobjargproc)Vector4f_ass_slice,	// sq_ass_slice
};

PyTypeObject Vector4fType = {
	PyObject_HEAD_INIT(NULL)

	0,								// ob_size
	"pyeigen.Vector4f",				// tp_name
	sizeof(Vector4f),				// tp_basicsize
	0,								// tp_itemsize
	0,								// tp_dealloc
	0,								// tp_print
	0,								// tp_getattr
	0,								// tp_setattr
	0,								// tp_compare
	(reprfunc)Vector4f_repr,		// tp_repr
	&Vector4f_numbermethods,		// tp_as_number
	&Vector4f_sequencemethods,		// tp_as_sequence
	0,								// tp_as_mapping
	0,								// tp_hash
	0,								// tp_call
	0,								// tp_str
	0,								// tp_getattro
	0,								// tp_setattro
	0,								// tp_as_buffer
	Py_TPFLAGS_DEFAULT |			// tp_flags
	Py_TPFLAGS_BASETYPE |
	Py_TPFLAGS_CHECKTYPES,
	0,								// tp_doc
	0,								// tp_traverse
	0,								// tp_clear
	0,								// tp_richcompare
	0,								// tp_weaklistoffset
	0,								// tp_iter
	0,								// tp_iternext
	Vector4f_methods,				// tp_methods
	0,								// tp_members
	Vector4f_getset,				// tp_getset
	0,								// tp_base
	0,								// tp_dict
	0,								// tp_descr_get
	0,								// tp_descr_set
	0,								// tp_dictoffset
	(initproc)Vector4f_init,		// tp_init
	0,								// tp_alloc
	Vector4f_new,					// tp_new
	0,								// tp_free
	0,								// tp_is_gc
	0,								// tp_bases
	0,								// tp_mro
	0,								// tp_cache
	0,								// tp_subclasses
	0,								// tp_weaklist
	0,								// tp_del
	0,								// tp_version_tag
};

// Helpers
int Vector4f_Check(PyObject* p)
{
	return p->ob_type == &Vector4fType;
}

// Construction
PyObject* Vector4f_new(PyTypeObject* type, PyObject* args, PyObject* kwds)
{
	return type->tp_alloc(type, 0);
}

int Vector4f_init(Vector4f* self, PyObject* args, PyObject* kwds)
{
	if(!Vector4f_set(self, args))
		return -1;

	return 0;
}

// Properties
PyObject* Vector4f_get_x(Vector4f* self, void* closure)
{
	return PyFloat_FromDouble(self->vector.x());
}

int Vector4f_set_x(Vector4f* self, PyObject* value, void* closure)
{
	self->vector.x() = (float)PyFloat_AsDouble(value);
	if(PyErr_Occurred() != NULL)
		return -1;

	return 0;
}

PyObject* Vector4f_get_y(Vector4f* self, void* closure)
{
	return PyFloat_FromDouble(self->vector.y());
}

int Vector4f_set_y(Vector4f* self, PyObject* value, void* closure)
{
	self->vector.y() = (float)PyFloat_AsDouble(value);
	if(PyErr_Occurred() != NULL)
		return -1;

	return 0;
}

PyObject* Vector4f_get_z(Vector4f* self, void* closure)
{
	return PyFloat_FromDouble(self->vector.z());
}

int Vector4f_set_z(Vector4f* self, PyObject* value, void* closure)
{
	self->vector.z() = (float)PyFloat_AsDouble(value);
	if(PyErr_Occurred() != NULL)
		return -1;

	return 0;
}

PyObject* Vector4f_get_w(Vector4f* self, void* closure)
{
	return PyFloat_FromDouble(self->vector.w());
}

int Vector4f_set_w(Vector4f* self, PyObject* value, void* closure)
{
	self->vector.w() = (float)PyFloat_AsDouble(value);
	if(PyErr_Occurred() != NULL)
		return -1;

	return 0;
}

PyObject* Vector4f_get_norm(Vector4f* self, void* closure)
{
	return PyFloat_FromDouble(self->vector.norm());
}

PyObject* Vector4f_get_norm_sq(Vector4f* self, void* closure)
{
	return PyFloat_FromDouble(self->vector.squaredNorm());
}

PyObject* Vector4f_get_normalized(Vector4f* self, void* closure)
{
	Vector4f* result = PyObject_New(Vector4f, &Vector4fType);
	if(result != NULL)
		result->vector = self->vector.normalized();

	return (PyObject*)result;
}

// Methods
PyObject* Vector4f_zero(PyTypeObject* cls, PyObject* noargs)
{
	Vector4f* result = PyObject_New(Vector4f, cls);
	if(result != NULL)
		result->vector.setZero();

	return (PyObject*)result;
}

PyObject* Vector4f_ones(PyTypeObject* cls, PyObject* noargs)
{
	Vector4f* result = PyObject_New(Vector4f, cls);
	if(result != NULL)
		result->vector.setOnes();

	return (PyObject*)result;
}

PyObject* Vector4f_identity(PyTypeObject* cls, PyObject* noargs)
{
	Vector4f* result = PyObject_New(Vector4f, cls);
	if(result != NULL)
		result->vector.setIdentity();

	return (PyObject*)result;
}

PyObject* Vector4f_constant(PyTypeObject* cls, PyObject* args)
{
	float constant = 0.0f;
	if(!PyArg_ParseTuple(args, "f", &constant))
		return NULL;

	Vector4f* result = PyObject_New(Vector4f, cls);
	if(result != NULL)
		result->vector.setConstant(constant);

	return (PyObject*)result;
}

PyObject* Vector4f_random(PyTypeObject* cls, PyObject* noargs)
{
	Vector4f* result = PyObject_New(Vector4f, cls);
	if(result != NULL)
		result->vector.setRandom();

	return (PyObject*)result;
}

PyObject* Vector4f_unit_x(PyTypeObject* cls, PyObject* noargs)
{
	Vector4f* result = PyObject_New(Vector4f, cls);
	if(result != NULL)
		result->vector << 1.f, 0.f, 0.f, 0.f;

	return (PyObject*)result;
}

PyObject* Vector4f_unit_y(PyTypeObject* cls, PyObject* noargs)
{
	Vector4f* result = PyObject_New(Vector4f, cls);
	if(result != NULL)
		result->vector << 0.f, 1.f, 0.f, 0.f;

	return (PyObject*)result;
}

PyObject* Vector4f_unit_z(PyTypeObject* cls, PyObject* noargs)
{
	Vector4f* result = PyObject_New(Vector4f, cls);
	if(result != NULL)
		result->vector << 0.f, 0.f, 1.f, 0.f;

	return (PyObject*)result;
}

PyObject* Vector4f_unit_w(PyTypeObject* cls, PyObject* noargs)
{
	Vector4f* result = PyObject_New(Vector4f, cls);
	if(result != NULL)
		result->vector << 0.f, 0.f, 0.f, 1.f;

	return (PyObject*)result;
}

PyObject* Vector4f_set(Vector4f* self, PyObject* args)
{
	float x = 0.0f;
	float y = 0.0f;
	float z = 0.0f;
	float w = 0.0f;
	if(!PyArg_ParseTuple(args, "|ffff", &x, &y, &z, &w))
		return NULL;

	self->vector << x, y, z, w;
	Py_RETURN_NONE;
}

PyObject* Vector4f_set_zero(Vector4f* self, PyObject* noargs)
{
	self->vector.setZero();
	Py_RETURN_NONE;
}

PyObject* Vector4f_set_ones(Vector4f* self, PyObject* noargs)
{
	self->vector.setOnes();
	Py_RETURN_NONE;
}

PyObject* Vector4f_set_constant(Vector4f* self, PyObject* args)
{
	float constant = 0.0f;
	if(!PyArg_ParseTuple(args, "f", &constant))
		return NULL;

	self->vector.setConstant(constant);
	Py_RETURN_NONE;
}

PyObject* Vector4f_set_random(Vector4f* self, PyObject* noargs)
{
	self->vector.setRandom();
	Py_RETURN_NONE;
}

PyObject* Vector4f_dot(Vector4f* self, PyObject* other)
{
	if(!Vector4f_Check(other))
	{
		PyErr_SetString(PyExc_ValueError, "Need Vector4f for dot product");
		return NULL;
	}

	Vector4f* otherv = (Vector4f*)other;
	return PyFloat_FromDouble(self->vector.dot(otherv->vector));
}

PyObject* Vector4f_normalize(Vector4f* self, PyObject* noargs)
{
	self->vector.normalize();
	Py_RETURN_NONE;
}

// Number methods
PyObject* Vector4f_add(PyObject* o1, PyObject* o2)
{
	if(!Vector4f_Check(o1) || !Vector4f_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector4f* v1 = (Vector4f*)o1;
	Vector4f* v2 = (Vector4f*)o2;
	Vector4f* result = PyObject_New(Vector4f, &Vector4fType);
	if(result != NULL)
		result->vector = v1->vector + v2->vector;

	return (PyObject*)result;
}

PyObject* Vector4f_subtract(PyObject* o1, PyObject* o2)
{
	if(!Vector4f_Check(o1) || !Vector4f_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector4f* v1 = (Vector4f*)o1;
	Vector4f* v2 = (Vector4f*)o2;
	Vector4f* result = PyObject_New(Vector4f, &Vector4fType);
	if(result != NULL)
		result->vector = v1->vector - v2->vector;

	return (PyObject*)result;
}

PyObject* Vector4f_multiply(PyObject* o1, PyObject* o2)
{
	Vector4f* v;
	float scalar;
	if(Vector4f_Check(o1) && PyNumber_Check(o2))
	{
		v = (Vector4f*)o1;
		scalar = (float)PyFloat_AsDouble(o2);
		if(PyErr_Occurred() != NULL)
			return NULL;

	}
	else if(PyNumber_Check(o1) && Vector4f_Check(o2))
	{
		scalar = (float)PyFloat_AsDouble(o1);
		v = (Vector4f*)o2;
		if(PyErr_Occurred() != NULL)
			return NULL;
	}
	else
		RETURN_NOTIMPLEMENTED;

	Vector4f* result = PyObject_New(Vector4f, &Vector4fType);
	if(result != NULL)
		result->vector = v->vector * scalar;

	return (PyObject*)result;
}

PyObject* Vector4f_divide(PyObject* o1, PyObject* o2)
{
	if(!Vector4f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector4f* v = (Vector4f*)o1;
	float scalar = (float)PyFloat_AsDouble(o2);
	if(PyErr_Occurred() != NULL)
		return NULL;

	Vector4f* result = PyObject_New(Vector4f, &Vector4fType);
	if(result != NULL)
		result->vector = v->vector / scalar;

	return (PyObject*)result;
}

PyObject* Vector4f_negative(PyObject* o)
{
	if(!Vector4f_Check(o))
		RETURN_NOTIMPLEMENTED;

	Vector4f* v = (Vector4f*)o;
	Vector4f* result = PyObject_New(Vector4f, &Vector4fType);
	if(result != NULL)
		result->vector = -v->vector;

	return (PyObject*)result;
}

PyObject* Vector4f_inplace_add(PyObject* o1, PyObject* o2)
{
	if(!Vector4f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector4f* v1 = (Vector4f*)o1;
	Vector4f* v2 = (Vector4f*)o2;
	v1->vector += v2->vector;

	return o1;
}

PyObject* Vector4f_inplace_subtract(PyObject* o1, PyObject* o2)
{
	if(!Vector4f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector4f* v1 = (Vector4f*)o1;
	Vector4f* v2 = (Vector4f*)o2;
	v1->vector -= v2->vector;

	return o1;
}

PyObject* Vector4f_inplace_multiply(PyObject* o1, PyObject* o2)
{
	Vector4f* v;
	float scalar;
	if(Vector4f_Check(o1) && PyNumber_Check(o2))
	{
		v = (Vector4f*)o1;
		scalar = (float)PyFloat_AsDouble(o2);
		if(PyErr_Occurred() != NULL)
			return NULL;
	}
	else if(PyNumber_Check(o1) && Vector4f_Check(o2))
	{
		scalar = (float)PyFloat_AsDouble(o1);
		v = (Vector4f*)o2;
		if(PyErr_Occurred() != NULL)
			return NULL;
	}
	else
		RETURN_NOTIMPLEMENTED;

	v->vector *= scalar;

	return (PyObject*)v;
}

PyObject* Vector4f_inplace_divide(PyObject* o1, PyObject* o2)
{
	if(!Vector4f_Check(o1) || !PyNumber_Check(o2))
		RETURN_NOTIMPLEMENTED;

	Vector4f* v = (Vector4f*)o1;
	float scalar = (float)PyFloat_AsDouble(o2);
	if(PyErr_Occurred() != NULL)
		return NULL;

	v->vector /= scalar;

	return (PyObject*)v;
}
// Sequence methods
Py_ssize_t Vector4f_length(Vector4f* self)
{
	return 4;
}

PyObject* Vector4f_item(Vector4f* self, Py_ssize_t index)
{
	if(index < 0 || index >= 4)
	{
		PyErr_SetString(PyExc_IndexError, "index out of range");
		return NULL;
	}

	return PyFloat_FromDouble(self->vector[index]);
}

PyObject* Vector4f_slice(Vector4f* self, Py_ssize_t index1, Py_ssize_t index2)
{
	if(index1 > 4)
		index1 = 4;
	if(index2 > 4)
		index2 = 4;
	if(index1 > index2)
		index1 = index2;

	Py_ssize_t length = index2 - index1;
	PyObject* tuple = PyTuple_New(length);
	if(tuple != NULL)
		for(Py_ssize_t index = 0; index < length; ++index)
		{
			PyObject* item = PyFloat_FromDouble(self->vector[index1 + index]);
			if(item == NULL)
			{
				Py_DECREF(tuple);
				return NULL;
			}

			PyTuple_SetItem(tuple, index, item);
		}

	return tuple;
}

int Vector4f_ass_item(Vector4f* self, Py_ssize_t index, PyObject* value)
{
	if(index < 0 || index >= 4)
	{
		PyErr_SetString(PyExc_IndexError, "index out of range");
		return NULL;
	}

	self->vector[index] = (float)PyFloat_AsDouble(value);
	if(PyErr_Occurred() != NULL)
		return -1;

	return 0;
}

int Vector4f_ass_slice(Vector4f* self, Py_ssize_t index1, Py_ssize_t index2,
	PyObject* value)
{
	if(index1 > 4)
		index1 = 4;
	if(index2 > 4)
		index2 = 4;
	if(index1 > index2)
		index1 = index2;

	Py_ssize_t length = index2 - index1;

	if(value == NULL || PySequence_Length(value) != length)
	{
		PyErr_SetString(PyExc_ValueError, "Can't change size of Vector4f");
		return -1;
	}

	for(Py_ssize_t index = index1; index < index2; ++index)
	{
		self->vector[index] =
			(float)PyFloat_AsDouble(PySequence_GetItem(value, index));
		if(PyErr_Occurred() != NULL)
			return -1;
	}

	return 0;
}

// Special methods
PyObject* Vector4f_repr(Vector4f* self)
{
	char buffer[1024];
	_snprintf(buffer, 1024, "Vector4f(%f, %f, %f, %f)", self->vector.x(),
		self->vector.y(), self->vector.z(), self->vector.w());
	return PyString_FromString(buffer);
}
